/*
** Module   : ComboBoxGuiImpl.java
** Abstract : GUI combo-box implementation
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 EVL 20150421 Initial version.
** 002 GES 20150508 Eliminated AWT dependencies.
** 003 EVL 20150514 Adding GUI specific keyboard and mouse default exit events.  Turning on
**                  tooltip engine.  Adding consistent cleanup on drop-down exit for GUI mode.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * Specific GUI implementation of COMBO-BOX widget.
 */
@SuppressWarnings("unchecked")
public class ComboBoxGuiImpl
extends ComboBox<GuiOutputManager>
{
   /** Shift in X to draw text inside entry. */
   public static final int TEXT_SHIFT = 4;
   
   /** Shift in X and Y to draw entry field for DROP-DOWN and SIMPLE modes. */
   public static final int ENTRY_SHIFT = 2;
   
   /** Dimension delta to draw entry field for DROP-DOWN and SIMPLE modes. */
   public static final int ENTRY_DELTA = ENTRY_SHIFT + 1;
   
   /** Screen driver */
   private GuiDriver gd;

   /** Output manager */
   private GuiOutputManager screen;
   
   /** Coordinate converter */
   private CoordinatesConversion cc;

   /** Color resolver */
   private GuiColorResolver gc;
   
   /** Font resolver */
   private GuiFontResolver gf;
      
   /** Selected item box color. */
   private ColorRgb selectionBack;

   /** Selected dotted line color. */
   private ColorRgb selectionDots;

   /** Light shadow box color. */
   private ColorRgb lightShadow;

   /** Dark shadow box color. */
   private ColorRgb darkShadow;
   
   /** Very dark shadow box color. */
   private ColorRgb veryDarkShadow;
   
   /** Pressed button state flag. */
   private boolean pressed = false;
   
   /** Image to draw drop down button. */
   private ImageGuiImpl imgButt = null; 
   
   /** Image to draw drop down button in disabled state. */
   private ImageGuiImpl imgButtDisabled = null;
   
   /** Editable entry field for SIMPLE and DROP-DOWN modes. */
   private FillInGuiImpl entryField = null;
   
   /** Name of the image to use for drop-down list activation button. */
   private static final String DROP_DOWN_BUTTON_NAME = "btn-down-arrow";
   
   /** Name of the image to use for drop-down list activation button disabled sate. */
   private static final String DROP_DOWN_BUTTON_NAME_INS = "btn-down-arrow-ins";
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public ComboBoxGuiImpl(WidgetId id, WidgetConfig cfg)
   {
      super(id, (ComboBoxConfig) cfg);
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    wcfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, ComboBoxConfig wcfg)
   {
      super.initialize(id, wcfg);
      
      // screen driver
      gd = (GuiDriver) OutputManager.getDriver();
      
      // output manager
      screen = screen();
      cc = screen.coordinates();
      
      // font resolver
      gf = new GuiFontResolver(Window.resolveWindow(this), gd, config(), false);
      
      // initialize the shadow colors
      lightShadow    = gd.getSysColor(SysColor.COLOR_WINDOW);
      darkShadow     = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      veryDarkShadow = gd.getSysColor(SysColor.COLOR_3DDKSHADOW);
      selectionBack  = gd.getSysColor(SysColor.COLOR_HIGHLIGHT);
      
      // the color below is negative to selectionBack
      selectionDots  = new ColorRgb(255 - selectionBack.getRed(),
                                    255 - selectionBack.getGreen(),
                                    255 - selectionBack.getBlue());

      // create required images to draw drop-down button
      imgButt = initDropDownButton(DROP_DOWN_BUTTON_NAME);
      imgButtDisabled = initDropDownButton(DROP_DOWN_BUTTON_NAME_INS);
      
      // register tooltip action listener
      gd.registerTooltipWorker(this, gf.font().font, ToolTip.isActive(config().tooltip));
   }

   /**
    * Draw the combo-box. Called by this ComboBox's parent container.
    */
   @Override
   public void draw()
   {
      // initialize entry field for SIMPLE and DROP-DOWN modes but only on the first draw
      if ((config().mode == ComboBoxConfig.Mode.SIMPLE ||
           config().mode == ComboBoxConfig.Mode.DROP_DOWN) && entryField == null)
      {
         // create new fill-in
         createEntryField();
         
         // and set up initial value for entry field
         entryField.setValue(new character(model().selected()));
      }
      
      // combo-box options need to be refreshed before every painting
      final ComboBoxConfig cbbcfg = config();
      gc = new GuiColorResolver(gd, cbbcfg, SysColor.COLOR_MENU, SysColor.COLOR_BTNTEXT);
      
      // determine colors to use
      gc.colors();
      
      // resolve font
      final String text = getLabelText();
      final FontCache fntCache = gf.font();

      // get location and size
      NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();

      // set up clipping rectangle
      NativeRectangle clip = clipRectangle(p, new NativeDimension(d.width, d.height));
      
      // clippable drawing start 
      gd.draw(p, clip, new Runnable()
      {

         @Override
         public void run()
         {
            // inside this call widget origin is always (0,0)
            NativePoint po = new NativePoint(0, 0);

            boolean focused = hasFocus() && isEnabled();

            // the field border            
            drawFieldBorder(po, d);
            // text in respective state
            if (entryField == null)
            {
               drawEntryField(po, d, fntCache.font, focused);
            }
            else // this means the mode is SIMPLE or DROP-DOWN
            {
               entryField.draw(null);
            }
            // and drop-down button
            drawButton(po, d, pressed);
         }
         
      });
   }

   
   /**
    * Sets position of terminal's cursor to the first char of the combo-box.
    */
   @Override
   public void drawCaret()
   {
      // this check is required since the method can be called from multiple
      // drivers
      
      // nothing to do in GUI
   }

   /**
    * Obtain the input focus.
    */
   @Override
   public void requestFocus()
   {
      if (isEnabled())
      {
         // generate the FOCUS_GAINED event
         if (entryField != null)
         {
            entryField.requestFocus();
         }
         else
         {
            super.requestFocus();
         }
      }
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.  Opening drop-down list.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      // open drop-down if not already opened
      if (isEnabled() && !dropDownActive)
      {
         // activate drop-down
         activate();
      }
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.  Redraw combo-box in pressed
    * state and request focus.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      // push button effect
      if (isEnabled() && !pressed)
      {
         requestFocus();
         pressed = true;
         repaint();
      }
   }

   /**
    * Notification of a mouse released event occurred for this widget.  Redraw combo-box in
    * released state.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      // release button effect
      if (isEnabled() && pressed)
      {
         pressed = false;
         repaint();
      }
   }

   /**
    * Process key events that have been generated by this component.
    * 
    * @param    keyEvent
    *           The key event to be handled.
    */
   @Override
   public void processKeyEvent(KeyInput keyEvent)
   {
      if (isEnabled() && !dropDownActive)
      {
         if (entryField != null && keyEvent.actionCode() != Keyboard.SE_VALUE_CHANGED)
         {
            entryField.processKeyEvent(keyEvent);
         }
         else
         {
            int newIndx = ScrollableList.getFirstCharSelection(keyEvent, model(),
                                                         model().getSelectedIndex());
            if (newIndx >= 0)
            {
               // select the new item
               setNewSelection(newIndx);
               // and consume event
               keyEvent.consume();
            }
            else
            {
               super.processKeyEvent(keyEvent);
            }
         }
      }
   }
   
   /**
    * Process the action associated with the given key in an independent drawing bracket.
    * <p>
    * This must be called only in the context of widget-level event processing.
    * 
    * @param    key
    *           The key to be processed.
    */
   @Override
   public void processSystemKey(final KeyInput key)
   {
      if (key.actionCode() == Keyboard.KA_CURSOR_DOWN)
      {
         // select the next item
         setNewSelection(model().getSelectedIndex() + 1);
      }
      else if (key.actionCode() == Keyboard.KA_CURSOR_UP)
      {
         // select the previous item
         setNewSelection(model().getSelectedIndex() - 1);
      }
      else
      {
         super.processSystemKey(key);
      }
   }

   /**
    * Calculate width limit.  In GUI we need to adjust the width for different font metrics.
    * 
    * @param   body
    *          Perform calculation for body part of widget.
    * 
    * @return  Calculated width limit.
    */
   @Override
   public double calcLimit(boolean body)
   {
      // this is the ChUI native limit
      double txtLimit = super.calcLimit(body);
      
      // calculate the GUI version
      // TODO: identify exactly how to compute this difference without hardcoding
      txtLimit += 6.0;
      
      return txtLimit;
   }

   /**
    * Get widget screen value.
    * 
    * @return  Current widget screen value.
    */
   @Override
   public BaseDataType getScreenValue()
   {
      return entryField != null ? new character(entryField.getScreenValue(true))
                                : super.getScreenValue();
   }

   /**
    * Posts value changed event to the queue.  This should be done in GUI mode after any changes
    * in entry field for SIMLPLE and DROP_DOWN combo-box modes.
    */
   public void entryFieldChanged(String newValue)
   {
      // put an ActionEvent onto the system event queue
      EventManager.postEvent(new ActionEvent(this, newValue));
      
      // honor any Progress triggers
      KeyInput vchng = new KeyInput(this, EventType.KEY_PRESSED, Keyboard.SE_VALUE_CHANGED, true);
      EventManager.postEvent(vchng);
      
      ConfigHelper.setModified(config, true);
   }
   
   /**
    * Posts value changed event to the queue.  This should be done in ChUI and GUI mode after
    * closing drop-down or in GUI mode after pressing UP/DOWN keys with focused combo-box.
    */
   @Override
   protected void valueChangedCommon()
   {
      // set up entry field if exists
      if (entryField != null)
      {
         entryField.setValue(new character(model().selected()));
      }
      
      // and call superclass
      super.valueChangedCommon();
   }
   
   /**
    * Exit the drop-down (the "selection" mode of the combo-box).  GUI specific cleanup.
    */
   @Override
   protected void exitDropDown()
   {
      // call generic cleanup from the superclass
      super.exitDropDown();
      // clean up registry from drop-down widgets added in GUI code
      deregisterDropdownWidgets(dropDown);
   }
   
   /**
    * Initializes the event set to return.  GUI specifics,
    * 
    * @return  The keyboard event to be used to exit drop-down.
    */
   @Override
   protected Set<String> initEventsSet()
   {
      // call common version
      Set<String> setGuiResult = super.initEventsSet();

      // add GUI specific messages
      setGuiResult.add("TAB");
      setGuiResult.add("LEAVE");
      
      return setGuiResult;
   }
   
   /**
    * Adding mouse clicks as exit events.  GUI specifics.
    * 
    * @param   listToChange
    *          The event list to modify.
    */
   @Override
   protected void addMouseExitEvent(EventList listToChange)
   {
      // call common version first
      super.addMouseExitEvent(listToChange);

      // mouse click on anywhere dismisses drop-down
      listToChange.addEvent("MOUSE-MENU-DOWN", true);
      listToChange.addEvent("MOUSE-EXTEND-DOWN", true);

      // enable drop-down painting
      ((DropDownGuiImpl)dropDown).enableDraw();
   }

   /**
    * Shows or hides the combo-box and its label.
    *
    * @param    visible_
    *           <code>true</code> to make the widget visible, 
    *           <code>false</code> to hide the widget.
    */
   @Override
   public void setVisible(boolean visible_)
   {
      super.setVisible(visible_);
   }
   
   /**
    * Clean up the main registry pool from drop-down widgets when it is closing, if the widget is
    * a container, make recursive call.
    */
   private void createEntryField()
   {
      // get ID for new widget
      WidgetId finId = WidgetId.nextID();

      // create config for fill-in widget
      FillInConfig finCfg = new FillInConfig(finId);

      // inherit some options from combo-box config
      finCfg.labels = false;
      finCfg.frameId = config.frameId;
      finCfg.dataType = config.dataType;
      finCfg.setDynamicFormat(config.format);

      // reconstruct widget, initialize and add to the registry
      WidgetRegistry reg = GuiOutputManager.instance().getRegistry(); 
      reg.reconstructWidget(finCfg);

      // set up the pointer to created entry field
      entryField = (FillInGuiImpl) reg.getComponent(finId);

      // and set correct location
      NativePoint cbbLoc = physicalLocation();
      entryField.setPhysicalLocation(cbbLoc.x + ENTRY_SHIFT, cbbLoc.y + ENTRY_SHIFT);
      
      // inform fill-in to be used as part of the combo-box
      entryField.setComboBoxMode(true, this);

      // enable and make visible
      entryField.setVisible(true);
      entryField.setEnabled(true);

      // add the widget to the frame the combo-box belongs to
      Frame<GuiOutputManager> frCombo = (Frame<GuiOutputManager>)UiUtils.locateFrame(this);
      frCombo.addWidget(entryField);
   }
   
   /**
    * Clean up the main registry pool from drop-down widgets when it is closing, if the widget is
    * a container, make recursive call.
    * 
    * @param    widget
    *           The root widget.
    */
   private void deregisterDropdownWidgets(Widget<GuiOutputManager> widget)
   {
      WidgetConfig cfg = widget.config();
      if (cfg != null)
      {
         // remove only one hat has config
         screen().getRegistry().removeWidget(widget.getId());
      }

      if (widget instanceof AbstractContainer)
      {
         Widget<GuiOutputManager>[] widgets = 
            ((AbstractContainer<GuiOutputManager>) widget).widgets();

         for (Widget<GuiOutputManager> w : widgets)
         {
            deregisterDropdownWidgets(w);
         }
      }
   }

   /**
    * Changes the selected value without opening drop-down.
    * 
    * @param   indexNew
    *          The index of the new item to select.
    */
   private void setNewSelection(int indexNew)
   {
      // new value can not be out of the current range
      if (indexNew < 0 || indexNew >= model().size())
      {
         return;
      }
      
      // set up the model with new selection
      model().setSelectedIndex(indexNew);
      model().setSelectedId(indexNew + 1);
      model().select(model().get(indexNew));

      // and post value changed event
      ThinClient.getInstance().independentEventDrawingBracket(ComboBoxGuiImpl.this, new Runnable()
      {
         @Override
         public void run()
         {
            valueChangedCommon();
            repaint();
         }
      });
   }

   /**
    * Sets up initial image object for drop down button.
    * 
    * @param   name
    *          Image name for drop-down button.
    *
    * @return   Image object for the given name.
    */
   private ImageGuiImpl initDropDownButton(String name)
   {
      ImageGuiImpl imgResult = null;
      // get the image data for requested image
      byte[] imageData = ImageHelper.getPrepackagedImageData(name);
      // initialize the given image
      if (imageData != null)
      {
         imgResult = new ImageGuiImpl(WidgetId.nextID());
         imgResult.setImage(imageData);
         ((ImageConfig)imgResult.config()).transparent = true;
         // set up button location
         NativePoint location = imgResult.physicalLocation();
         // the size of the entry field
         NativeDimension dim = physicalDimension();
         // button is on the right corner of the entry field
         imgResult.setPhysicalLocation(location.x + dim.width - ButtonGuiImpl.PREBUILT_SIZE - 1,
                                       location.y + (dim.height - ButtonGuiImpl.PREBUILT_SIZE)/2);
      }
      return imgResult;
   }
   
   /**
    * Draw 3D border around the combo-box widget.
    * 
    * @param   orig
    *          Combo-box origin.
    * @param   dim
    *          Combo-box size.
    */
   private void drawFieldBorder(NativePoint orig, NativeDimension dim)
   {
      int x1 = orig.x;
      int x2 = orig.x + dim.width;
      int y1 = orig.y;
      int y2 = orig.y + dim.height;
      // drawing the frame around entry field
      
      gd.setColor(darkShadow);
      gd.drawLine(x1, y1, x1, y2);
      gd.drawLine(x1 + 1, y1, x2, y1);

      gd.setColor(gc.bgColor);
      gd.drawLine(x1 + 1, y2, x2, y2);
      gd.drawLine(x2, y2 - 1, x2, y1 + 1);

      gd.setColor(veryDarkShadow);
      gd.drawLine(x1 + 1, y1 + 1, x1 + 1, y2 - 1);
      gd.drawLine(x1 + 2, y1 + 1, x2 - 1, y1 + 1);
   }
   
   /**
    * Draw entry field of the combo-box widget.
    * 
    * @param   orig
    *          Combo-box origin.
    * @param   dim
    *          Combo-box size.
    * @param   fnt
    *          Font to use for combo-box entry field.
    * @param   focused
    *          Is the combo-box focused.
    */
   private void drawEntryField(NativePoint orig,
                               NativeDimension dim,
                               FontDetails<?> fnt,
                               boolean focused)
   {
      // coordinates of the field area
      int w = dim.width - ButtonGuiImpl.PREBUILT_SIZE - 4;
      int h = dim.height - 3; 

      // draw internal background
      gd.setColor(lightShadow);
      gd.drawRect(orig.x + 2, orig.y + 2, w, h );
      if (focused)
      {
         gd.setColor(selectionBack);
      }
      gd.fillRect(orig.x + 3, orig.y + 3, w - 1, h - 1);
      
      // get the text to draw with field
      String display = null;
      
      display = model().selected();
      
      if (display == null)
         display = new String("");

      // limit the data portion of the field to the maximum allowed size
      // in chui char coordinates are always whole numbers
      if (display.length() > (int) limit)
         display = display.substring(0, (int) limit);
      else if (display.length() < (int) limit)
         display = StringHelper.leftAlignText(display, (int) limit);

      // draw entry field text
      gd.setGuiFont(fnt);
      // left-center alignment
      int xt = orig.x + TEXT_SHIFT;
      int yt = orig.y + 1 + (dim.height + fnt.pointSize)/2;
      gd.setFontStyle(fnt.style);
      // enabled state controls text drawing
      if (isEnabled())
      {
         if (focused)
         {
            gd.setColor(lightShadow);
         }
         else
         {
            gd.setColor(gc.fgColor);
         }
         gd.drawString(display, xt, yt);
      }
      else
      {
         gd.setColor(lightShadow);
         gd.drawString(display, xt + 1, yt + 1);
         gd.setColor(darkShadow);
         gd.drawString(display, xt, yt);
      }

      // draw dotted rectangle around focused button text
      if (focused)
      {
         // set dotted line style
         gd.setLineStroke(LineStroke.DOTTED_SMALL);
         gd.setColor(selectionDots);
         gd.drawRect(orig.x + 3, orig.y + 3, w - 2, h - 2);
         // restore default line style
         gd.setLineStroke(LineStroke.DEFAULT);
      }
   }

   /**
    * Draw drop down button of the combo-box widget.
    * 
    * @param   orig
    *          Combo-box origin.
    * @param   dim
    *          Combo-box size.
    * @param   pressed
    *          Flag indicating the button state, pressed or released.
    */
   private void drawButton(NativePoint orig, NativeDimension dim, boolean pressed)
   {
      // simple check to prevent NPE
      if (imgButt == null || imgButtDisabled == null)
      {
         return;
      }
      
      // coordinates of the button area
      int x1 = orig.x + dim.width - ButtonGuiImpl.PREBUILT_SIZE - 1;
      int x2 = orig.x + dim.width;
      int y1 = orig.y + 2;
      int y2 = orig.y + dim.height;

      // draw the light frame around image
      gd.setColor(gc.bgColor);
      gd.fillRect(x1, y1, x2 - x1, y2 - y1);
      
      // disabled state has separate inage to draw  
      if (pressed)
      {
         // for pressed state need 1 pixel right-down shift 
         NativePoint location = imgButt.physicalLocation();
         imgButt.setPhysicalLocation(location.x + 1, location.y + 1);
         imgButt.draw();
         // draw small button frame
         gd.setColor(darkShadow);
         gd.drawRect(location.x, y1, x2 - x1 - 1, y2 - y1 - 1);
         // and restore origin after draw
         imgButt.setPhysicalLocation(location.x, location.y);
      }
      else
      {
         // determine what image to draw
         ImageGuiImpl imgReleased = null;
         if (isEnabled())
         {
            imgReleased = imgButt;
         }
         else
         {
            imgReleased = imgButtDisabled; 
         }
         imgReleased.draw();
         // draw small button frame
         NativePoint location = imgReleased.physicalLocation();
         gd.setColor(lightShadow);
         gd.drawLine(location.x + 1, y1 + 2,
                     location.x + 1, y2 - 3);
         gd.drawLine(location.x + 1, y1 + 1,
                     location.x + ButtonGuiImpl.PREBUILT_SIZE - 2, y1 + 1);

         gd.setColor(darkShadow);
         gd.drawLine(location.x + 1, y2 - 2,
                     location.x + ButtonGuiImpl.PREBUILT_SIZE - 1, y2 - 2);
         gd.drawLine(location.x + ButtonGuiImpl.PREBUILT_SIZE - 1, y1 + 1,
                     location.x + ButtonGuiImpl.PREBUILT_SIZE - 1, y2 - 3);

         gd.setColor(veryDarkShadow);
         gd.drawLine(location.x, y2 - 1,
                     location.x + ButtonGuiImpl.PREBUILT_SIZE, y2 - 1);
         gd.drawLine(location.x + ButtonGuiImpl.PREBUILT_SIZE, y1,
                     location.x + ButtonGuiImpl.PREBUILT_SIZE, y2 - 2);
      }
   }
}
