/* 
** Module   : FillInGuiImpl.java
** Abstract : text field editor/display implementation, GUI-specific version.
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20141220 Created initial version.
** 002 CA  20150323 Fixed calculations of widget dimensions in GUI.
** 003 HC  20150323 Removed business logic from config classes.
** 004 CA  20150406 Fixed fill-in height: is computed using the same formula as pixels-per-row, 
**                  but using the explicit font, if set.
** 005 CA  20150507 RETURN doesn't clear selection.
**                  Fixed text drawing - centered vertically.
** 006 CA  20150510 Fixed selection/border color and added a NPE protection.
** 007 GES 20150425 Eliminated AWT dependencies.
** 008 EVL 20150601 Adding specialmode to enable correct drawing inside combo-box.
*/

package com.goldencode.p2j.ui.client.gui;

// this usage of AWT events is intentional, they are being used separately from the
// AWT/Swing event model itself; no other AWT/Swing usage is occurring in this class
import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.driver.*;

/**
 * Implementation of Progress FILL-IN component for GUI.
 */
public class FillInGuiImpl
extends FillIn<GuiOutputManager,FillInConfig>
{
   /** Screen driver */
   private GuiDriver gd;
   
   /** Font resolver */
   private GuiFontResolver gf;

   /**
    * Flag indicating initialization is not needed on a focus-gained event, as it was treated
    * explicitly by some other logic (like mouse clicked).
    */
   private boolean bypassFocusInit = false;
   
   /** Flag indicating the fill-in is the part of the combo-box widget. */
   private boolean comboBoxMode = false;
   
   /** Reference to the possible combo-box object. */
   private ComboBoxGuiImpl comboBoxRef = null;

   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public FillInGuiImpl(WidgetId id, FillInConfig cfg)
   {
      super(id, cfg);
      
      this.nativeInsets = new NativeInsets(2, 2, 2, 2);
   }

   /**
    * Compute the widget's width in native units.
    * 
    * @param    widget
    *           The FILL-IN or TEXT widget.
    * @param    gd
    *           The GUI driver in use.
    * @param    gf
    *           The font resolver.
    * @param    isText
    *           <code>true</code> if this is for a TEXT widget.
    *           
    * @return   The widget's width in native units.
    */
   protected static int nativeWidth(FillIn<GuiOutputManager,?> widget, 
                                    GuiDriver                  gd,
                                    GuiFontResolver            gf,
                                    boolean                    isText)
   {
      Window<?> window = Window.resolveWindow(widget);
      FillInConfig config = widget.config();
   
      if (gf == null || config == null || config.appDataPres == null)
         return 0;
   
      boolean explicit = (config.font != FontManager.UNDEFINED_FONT); 
      boolean fixedFontFormat = config.appDataPres.isFixedFont();
      int font = gf.resolveFontNum();
      int fmtWidth = widget.isAtField() ? config.atFormatLength : config.appFormat.getScreenWidth();
      int avgFontWidth = FontManager.getFontWidth(window, font, gd);
      boolean fixedFont = false;
      if (!fixedFontFormat && !explicit)
      {
         fixedFont = FontManager.isFixedSize(window, font, gd);
      }
      int maxFontWidth = FontManager.getMaxFontWidth(window, font, gd);

      // In GUI, the width has a deviation from the normal "multiply format width with the font's 
      // average character width" formula.  After gathering fill-in metrics for various fonts and 
      // format lengths, it was determined that a special formula is used when all these are true:
      // - the format length is between 1 and 12 characters
      // - the font is not set explicitly via the FONT phrase 
      // - the widget normally expects a fixed font (like the case for a fill-in with a numeric 
      //   value)
      // A java program used to prove this formula can be found in the testcases project, 
      // the uast/fillin/CheckMetricsFillin.java file.
      
      int pixelWidth = !fixedFontFormat && !explicit && fmtWidth >= 1 && fmtWidth <= 12
                          // handle font sizes 1 to 12 
                          ? Math.min(fmtWidth, 3) * 
                           (fixedFont && !explicit ? maxFontWidth 
                                                    : (avgFontWidth * 2 + 
                                                       (int) Math.floor(avgFontWidth / 2))) +  
                           Math.max(fmtWidth - 3, 0) * avgFontWidth
   
                           // default case
                           : fmtWidth * avgFontWidth;
      
      return pixelWidth + (isText ? 0 : 8 + (FontManager.isThreeD() ? 2 : 0));
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, FillInConfig cfg)
   {
      super.initialize(id, cfg);

      gd = (GuiDriver) OutputManager.getDriver();
      Window<?> window = Window.resolveWindow(this);
      boolean fixedFont = config.appDataPres.isFixedFont();
      gf = new GuiFontResolver(window, gd, config, fixedFont);
      updateSize();

      addFocusListener(new FocusListener()
      {
         @Override
         public void onFocusLost(FocusEvent event)
         {
            invalidateSelection();
            
            resetEditScroller();
         }
         
         @Override
         public void onFocusGained(FocusEvent event)
         {
            if (bypassFocusInit)
            {
               bypassFocusInit = false;
               return;
            }

            int lastOffset = getLastOffset();
            
            if (getText(true).length() > 0)
            {
               // select entire text
               config.selectionStart = 0;
               config.selectionEnd = lastOffset + 1;
               moveCursorToLast();
            }
            else
            {
               invalidateSelection();
               setCursorOffset(0);
            }
         }
      });
   }
   
   /**
    * Retrieve the widget's width in native units.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeWidth()
   {
      if (config.fixedWidth)
      {
         return config.widthPixels;
      }
      
      return nativeWidth(this, gd, gf, false);
   }
   
   /**
    * Retrieve the widget's height in native units.
    * 
    * @return   See above.
    */
   @Override
   protected int nativeHeight()
   {
      if (config.fixedHeight)
      {
         return config.heightPixels;
      }

      if (gf == null || config == null || config.appDataPres == null)
         return 0;

      Window<?> window = Window.resolveWindow(this);
      int font = gf.resolveFontNum();
      if (font == FontManager.DEFAULT_FIXED_FONT)
      {
         // fixed-font is used only for width
         font = FontManager.DEFAULT_FONT;
      }
      boolean threeD = FontManager.isThreeD();

      int legacyHeight = FontManager.getFontHeight(window, font, screen().getInstanceDriver());
      legacyHeight = legacyHeight + (int) (legacyHeight / 2) + (threeD ? 2 : 0);

      return legacyHeight;
   }

   /**
    * Returns the last valid cursor position from the left bound of the FillIn.
    * 
    * @return  The last valid cursor offset.
    */
   @Override
   public int getLastOffset()
   {
      return Math.max(getText(true).length() - 1, 0);
   }
   
   /**
    * Enable/disable the widget for editing.  During this operation, any data values are saved off 
    * for subsequent restore (during editing) using the <code>RECALL</code> key.
    * <p>
    * In GUI, it registers this widget with the driver, to allow responsive actions when the
    * mouse is hovered over the widget.
    * 
    * @param    enable
    *           <code>true</code> to enable editing in this widget.
    */
   @Override
   public void setEnabled(boolean enable)
   {
      boolean isEnabled = isEnabled();
      
      super.setEnabled(enable);
      
      int id = this.config.id.asInt();
      if (gd == null)
      {
         return;
      }

      if (enable)
      {
         if (!isEnabled)
         {
            gd.registerEditableWidget(id);
         }
      }
      else
      {
         if (isEnabled)
         {
            gd.deregisterWidget(id);
         }
      }
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.
    * <p>
    * If the widget is not already focused, it forces the focus to it.  After the focus was
    * received, it computes the caret location, based on the mouse position over the widget's
    * content.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      if (!isEnabled())
      {
         return;
      }
      
      if (!hasFocus())
      {
         ThinClient.getInstance().independentEventDrawingBracket(this, new Runnable()
         {
            @Override
            public void run()
            {
               bypassFocusInit = true;
               
               // process the focus here, so we can continue with the special processing bellow.
               requestFocus();
               
               invalidateSelection();
            }
         });
      }
      
      String stxt = getText();

      invalidateSelection();
      
      // determine which letter is hovered
      int x = textNativeLocation().translate(screenPhysicalLocation()).x;
      
      int idx = 1;
      FontCache fc = gf.font();
      FontDetails<?> font = fc.font;
      String txt = "";
      while (idx < stxt.length())
      {
         txt = stxt.substring(0, idx);

         int txtWidth = gd.getTextWidth(txt, font);
         
         if (e.getX() <= x + txtWidth)
         {
            break;
         }

         idx = idx + 1;
      }
      
      // hovered letter is at index "idx".  need to found which side to draw the caret
      if (idx == stxt.length())
      {
         // mouse is on the right-side of the right-most letter
         moveCursorToLast();
      }
      else if (txt.length() > 0)
      {
         // find where is the hovering - left-side or right-side
         int leftSideWidth = gd.getTextWidth(txt.substring(0, txt.length() - 1), font);
         int fullWidth = gd.getTextWidth(txt, font);
         int letterWidth = fullWidth - leftSideWidth;
         int midLetter = x + leftSideWidth + Math.round(letterWidth / 2);
         setCursorOffset(e.getX() >= midLetter ? txt.length() : txt.length() - 1);
      }
   
      repaint();
   }

   /**
    * Process GUI specific keys:
    * <ul>
    *    <li><code>VK_SHIFT_CURSOR_RIGHT</code> - expands text selection to the right.</li>
    *    <li><code>VK_SHIFT_CURSOR_LEFT</code> - expands text selection to the left.</li>
    *    <li><code>VK_SHIFT_HOME</code> - expands text selection to the first char.</li>
    *    <li><code>VK_SHIFT_END</code> - expands text selection to the last char.</li>
    *    <li><code>VK_CTRL_A</code> - selects the entire content.</li>
    *    <li><code>KA_COPY</code> - copies the selected text to the OS clipboard.</li>
    *    <li><code>KA_CUT</code> - copies and deletes the selected text.</li>
    *    <li><code>KA_PASTE</code> - inserts the OS cliboard content into the widget.</li>
    * </ul>
    * 
    * @param   ke
    *          Key event structure.
    */
   @Override
   public void processKeyEvent(KeyInput ke)
   {
      int key = ke.actionCode();
      boolean invalidateSelection = true;
      int lastOffset = getLastOffset();
      int offset = getCursorOffset();
      if (key == Key.VK_SHIFT_CURSOR_RIGHT)
      {
         // expand selection
         if (config.selectionStart == FillInConfig.INVALID_SELECTION)
         {
            config.selectionStart = offset;
            config.selectionEnd = Math.min(offset + 1, lastOffset + 1);
         }
         else if (offset == config.selectionStart)
         {
            config.selectionStart = Math.min(config.selectionStart + 1, lastOffset);
            
            if (config.selectionStart > config.selectionEnd)
            {
               int aux = config.selectionStart;
               config.selectionStart = config.selectionEnd;
               config.selectionEnd = aux;
            }
         }
         else if (config.selectionEnd != offset + 1)
         {
            config.selectionEnd = Math.min(offset + 1, lastOffset + 1);
         }
         else
         {
            config.selectionEnd = Math.min(config.selectionEnd + 1, lastOffset + 1);
         }
         
         ke.setKeyCode(Keyboard.KA_CURSOR_RIGHT);
         ke.setActionCode(Keyboard.KA_CURSOR_RIGHT);
         invalidateSelection = false;
      }
      else if (key == Key.VK_SHIFT_CURSOR_LEFT)
      {
         // expand selection
         if (config.selectionEnd == FillInConfig.INVALID_SELECTION)
         {
            config.selectionEnd = Math.min(offset, lastOffset + 1);
            config.selectionStart = Math.max(offset - 1, 0);
         }
         else if (offset == config.selectionEnd)
         {
            config.selectionEnd = Math.max(config.selectionEnd - 1, 0);
            
            if (config.selectionStart > config.selectionEnd)
            {
               int aux = config.selectionStart;
               config.selectionStart = config.selectionEnd;
               config.selectionEnd = aux;
            }
         }
         else if (config.selectionStart != offset)
         {
            config.selectionStart = Math.max(offset - 1, 0);
         }
         else
         {
            config.selectionStart = Math.max(config.selectionStart - 1, 0);
         }
         
         ke.setKeyCode(Keyboard.KA_CURSOR_LEFT);
         ke.setActionCode(Keyboard.KA_CURSOR_LEFT);
         invalidateSelection = false;
      }
      else if (key == Key.VK_SHIFT_HOME)
      {
         // expand selection
         if (config.selectionEnd == FillInConfig.INVALID_SELECTION)
         {
            config.selectionEnd =  Math.min(offset + 1, lastOffset + 1);
         }
         else if (offset == config.selectionEnd)
         {
            config.selectionEnd = config.selectionStart;
         }
         config.selectionStart = 0;
         
         ke.setKeyCode(Keyboard.KA_HOME);
         ke.setActionCode(Keyboard.KA_HOME);
         invalidateSelection = false;
      }
      else if (key == Key.VK_SHIFT_END)
      {
         // expand selection
         if (config.selectionStart == FillInConfig.INVALID_SELECTION)
         {
            config.selectionStart = offset;
         }
         else if (offset == config.selectionStart)
         {
            config.selectionStart = config.selectionEnd;
         }

         config.selectionEnd = lastOffset + 1;
         
         ke.setKeyCode(Keyboard.KA_END);
         ke.setActionCode(Keyboard.KA_END);
         invalidateSelection = false;
      }
      else if (key == Key.VK_CTRL_A)
      {
         if (getText(true).length() == 0)
         {
            invalidateSelection();
         }
         else
         {
            config.selectionStart = 0;
            config.selectionEnd = lastOffset + 1;
            repaint();
         }
         
         return;
      }
      else if (key == Keyboard.KA_COPY || key == Keyboard.KA_CUT)
      {
         if (validSelection())
         {
            String txt = getText(true);
            
            String selectedText = txt.substring(config.selectionStart, config.selectionEnd);
            
            // copy first
            gd.copyToCliboard(selectedText);

            // clear as needed
            if (key == Keyboard.KA_CUT)
            {
               clear(config.selectionStart, config.selectionEnd);
            }
         }
         
         ke.consume();
      }
      else if (key == Keyboard.KA_PASTE)
      {
         if (validSelection())
         {
            // clear everything first
            clear(config.selectionStart, config.selectionEnd);
         }
         
         // insert the text
         String txt = gd.clipboardContents();
         
         for (char ch : txt.toCharArray())
         {
            if (!config.appDataPres.input(ch))
            {
               // widget no longer accepts this...
               break;
            }
         }
         
         repaint();
         
         ke.consume();
      }
      
      key = ke.actionCode();
      
      if (invalidateSelection && validSelection())
      {
         // TODO: date format does not allow clearing...
         // TODO: date format does not allow BACKSPACE/DELETE
         if ((key >= 0x20 && key <= 0xff)        ||
              key == Keyboard.KA_BACKSPACE       ||
              key == Keyboard.KA_DELETE_CHARACTER)
         {
            // clear the selection
            clear(config.selectionStart, config.selectionEnd);
            
            if (key == Keyboard.KA_BACKSPACE || key == Keyboard.KA_DELETE_CHARACTER)
            {
               ke.consume();
            }
         }

         // no more selection
         if (key != Keyboard.KA_CLEAR)
         {
            // selection needs to survive, it will be used by the FillIn.processKeyEvent
            invalidateSelection();
         }
      }
      
      if (getCursorOffset() >= getCurrentFormat().getScreenWidth() && key >= 0x20 && key <= 0xff)
      {
         // can not add keys past the last char...
         return;
      }
      
      if (key == Keyboard.KA_END)
      {
         moveCursorToLast();
         adjustEditScroller(Keyboard.KA_CURSOR_RIGHT);
         refresh();
         
         return;
      }
      
      if (key == Keyboard.KA_HOME)
      {
         moveCursorToFirst();
      }
      
      
      if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         if (offset == lastOffset)
         {
            // this needs to be treated explicitly... otherwise the DisplayFormat classes will
            // not allow the cursor to get past the last position via cursor-right key
            setCursorOffset(offset + 1);
            
            // the cursor will remain on the current position, but will be drawn right-side
            adjustEditScroller(Keyboard.KA_CURSOR_RIGHT);
            if (editScroller == rightAnchor)
            {
               editScroller.setAnchorIndex(offset + 1);
            }
            refresh();
            
            return;
         }
      }
      if (key == Keyboard.KA_CURSOR_LEFT && isRightAlignedInt())
      {
         if (offset >= lastOffset + 1)
         {
            // this needs to be treated explicitly... otherwise the DisplayFormat classes will
            // not allow the cursor to get past the last position via cursor-right key
            setCursorOffset(lastOffset);
            
            refresh();
            
            return;
         }
      }
      
      // if combo-box mode is active and we have value changed in fill-in
      if (key == Keyboard.SE_VALUE_CHANGED && comboBoxMode)
      {
         // stop handling this event for fill-in
         ke.consume();
         // and reroute them to the enclosing combo-box
         comboBoxRef.entryFieldChanged(getText());
      }
      else
      {
         super.processKeyEvent(ke);
      }
      
      adjustEditScroller(key);
   }
   
   /**
    * Draws the caret.
    */
   @Override
   public void drawCaret()
   {
      // this check is required since the method can be called from multiple drivers
      if (!hasFocus() || !isEnabled())
      {
         return;
      }

      final FontCache fc = gf.font();
      final int cursorX = cursorPhysicalColumn();
      NativePoint eorig = this.screenPhysicalLocation().translate(editableNativeLocation());
      eorig.x--;
      
      NativeDimension edim = editableNativeDimension();
      edim.width = edim.width + nativeInsets.right;
      
      final NativeRectangle eclip = clipRectangle(eorig, edim);

      gd.drawFromOrigin(eorig, eclip, new Runnable()
      {
         @Override
         public void run()
         {
            gd.setColor(gd.getSysColor(SysColor.COLOR_CARET));
            gd.drawLine(cursorX, 0, cursorX, fc.fontHeight - 1);
         }
      });
   }

   /**
    * Actual drawing routine.
    *
    * @param   color
    *          Color for drawing.
    */
   @Override
   public void draw(Color color)
   {
      final GuiColorResolver gce = new GuiColorResolver(gd, config, 
                                                        SysColor.COLOR_HIGHLIGHT, 
                                                        SysColor.COLOR_HIGHLIGHTTEXT);
      final GuiColorResolver gcd = new GuiColorResolver(gd, config, 
                                                        SysColor.COLOR_FILLIN_BG,
                                                        SysColor.COLOR_FILLIN_FG);
      gcd.boxColor = gcd.sysFgColor;
      gce.selectionBgColor = gd.getSysColor(SysColor.COLOR_HOTLIGHT);
      gce.selectionFgColor = gd.getSysColor(SysColor.COLOR_HIGHLIGHTTEXT);
   
      // If uninitialized and not for input and not date fill-in (in this case
      // separators should be displayed), display blanks instead.
      // TODO: probably it should be replaced by variable initialization
      //       in FrameHelper.clearComponents().
      final boolean uninit = (getState() == ScreenBuffer.UNINITIALIZED)    && 
                              !(config.appFormat != null               && 
                                config.appFormat instanceof DateFormat &&
                                isEnabled());

      gce.colors();
      gcd.colors();
      
      NativePoint p = physicalLocation();
      final NativeDimension dim = physicalDimension();
      final String txt = getText(true);
      final NativeRectangle clip = clipRectangle(p, dim);
      final FontCache fc = gf.font();
      final FontDetails<?> font = fc.font;
      
      gd.draw(p, clip, new Runnable()
      {
         @Override
         public void run()
         {
            // clear the area
            gd.setColor(gcd.bgColor);
            gd.fillRect(0, 0, dim.width + 1, dim.height + 1);
            
            // draw the border
            if (isEnabled() && !comboBoxMode)
            {
               gd.setColor(gcd.boxColor);
               gd.drawRect(0, 0, dim.width, dim.height);
            }

            if (!uninit)
            {
               NativePoint txtLoc = textNativeLocation();
               final NativeDimension edim = editableNativeDimension();
               edim.height = gd.getTextHeight(txt, font) + 2;
               NativeRectangle txtClip = clipRectangle(txtLoc, edim);
               
               gd.draw(txtLoc, txtClip, new Runnable()
               {
                  @Override
                  public void run()
                  {
                     // draw full text (as if nothing is selected), first
                     gd.setGuiFont(font);
                     gd.setColor(gcd.fgColor);

                     int textHeight = gd.getTextHeight(txt, font);
                     final int ytxt;
                     
                     if (!isEnabled())
                     {
                        // needs to be centered vertically
                        gd.drawStringCentered(txt, 0, edim.height);
                     }
                     else
                     {
                        // needs to be aligned to top border
                        ytxt = textHeight;
                        gd.drawString(txt, 0, ytxt);
                     }

                     drawSelection();
                  }
               });
            }

            drawCaret();
         }

         private void drawSelection()
         {
            // draw the selection
            if (!hasFocus() || !isEnabled() || !validSelection())
            {
               return;
            }
   
            int txtHeight = gd.getTextHeight(txt, font);

            String leftSideText = txt.substring(0, config.selectionStart);
            String selectedText = txt.substring(config.selectionStart, config.selectionEnd);
            
            int leftSideSize = leftSideText.length() == 0 ? 0 : gd.getTextWidth(leftSideText, font);
            int seleSize = gd.getTextWidth(selectedText, font);
            int seleX = leftSideSize;
          
            // draw the background
            gd.setColor(gce.selectionBgColor);  
            gd.fillRect(seleX, 0, seleSize, txtHeight + 2);
            
            // draw the text
            gd.setColor(gce.selectionFgColor);
            gd.drawString(selectedText, seleX, txtHeight);
         }
      });
   }

   /**
    * Compute the initial cursor position, depending on the widget's implementation.
    * <p>
    * In GUI, the initial cursor position is always on the right-side text edge.
    *  
    * @return   See above.
    */
   @Override
   public int getInitialCursorPos()
   {
      return getText(true).length();
   }

   /**
    * Setting new value for combo-box mode flag.
    *
    * @param   mode
    *          <CODE>TRUE</CODE> if the fill-in is used as part of the combo-box,
    *          <CODE>FALSE</CODE> otherwise.
    * @param   cbbRef
    *          Reference to combo-box object the fill-in is the part of.
    */
   public void setComboBoxMode(boolean mode, ComboBoxGuiImpl cbbRef)
   {
      comboBoxMode = mode;
      comboBoxRef  = cbbRef;
   }

   /**
    * Get widget physical dimension, expressed in units native to UI type.
    *  
    * @return  widget dimension.
    */
   @Override
   public NativeDimension physicalDimension()
   {
      NativeDimension dim = super.physicalDimension();
      
      // if widget inside combo-box, need to decrease height
      if (comboBoxMode)
      {
         dim.height -= ComboBoxGuiImpl.ENTRY_DELTA;
      }
      
      return dim;
   }

   /**
    * Get widget width.
    * 
    * @return  Widget width.
    */
   @Override
   public double width()
   {
      double origWidth = super.width();
      
      // if widget inside combo-box, need to increase width
      if (comboBoxMode)
      {
         origWidth += 4.0;
      }
      
      return origWidth;
   }

   /**
    * Adjust the edit scroller depending on the key being processed and the cursor location.
    * 
    * @param    key
    *           The currently processed key.
    */
   @Override
   protected void adjustEditScroller(int key)
   {
      int offset = getCursorOffset();
      int last = getLastOffset();
      
      if (key == Keyboard.KA_DELETE_CHARACTER)
      {
         String txt = getText(true);
         FontCache fc = gf.font();
         FontDetails<?> font = fc.font;
         
         txt = txt.substring(0, offset);
         
         int tlen = gd.getTextWidth(txt, font);
         
         if (tlen > editableNativeDimension().width)
         {
            editScroller = rightAnchor;
            editScroller.setAnchorIndex(Math.min(offset + 1, last));
         }
         else
         {
            editScroller = leftAnchor;
            editScroller.setAnchorIndex(0);
         }
      }
      else
      {
         super.adjustEditScroller(key);
      }
   }

   /**
    * Get valid cursor position.
    * 
    * @return   valid cursor position.
    */
   protected int getValidCursorPos()
   {
      int pos = Math.min(getCursorOffset() + editScroller.getDelta(), getLastOffset() + 1);
      return pos;
   }
   
   /**
    * Get the cursor position, in native units, as it would be positioned in the fully-visible
    * text.
    * 
    * @return   See above.
    */
   @Override
   protected int getNativeCursorPos()
   {
      if (gd == null)
      {
         return 0;
      }

      int pos = getCursorOffset();
      
      String fullTxt = getText(true);
      String txt = pos >= fullTxt.length() ? fullTxt : fullTxt.substring(0, pos);

      FontCache fc = gf.font();
      FontDetails<?> font = fc.font;

      NativePoint np = new NativePoint(gd.getTextWidth(txt, font), 0);
      np = editScroller.adjustLocation(fullTxt, np);
      
      return np.x;
   }
   
   /**
    * Returns the insert mode state.
    * 
    * @return  In GUI, this is always <code>false</code>.
    */
   protected boolean getInsertMode()
   {
      return false;
   }
   
   /**
    * Reset the edit scroller to the initial state.
    */
   @Override
   protected void resetEditScroller()
   {
      boolean focused = hasFocus();
      if (editScroller == null || !focused)
      {
         // everything starts from the left-side
         editScroller = leftAnchor;
         editScroller.setAnchorIndex(0);
         
         if (!focused)
         {
            return;
         }
      }
      
      if (gd == null || gf == null)
      {
         // not yet initialized, nothing to do
         return;
      }
      
      int cursorPos = getInitialCursorPos();
      int txtWidth = gd.getTextWidth(getText(true), gf.font().font);
      boolean rightEndCursor = txtWidth >= editableNativeDimension().width;

      editScroller = (rightEndCursor ? rightAnchor : leftAnchor);

      editScroller.setAnchorIndex(rightEndCursor ? cursorPos : 0);
   }
   
   /**
    * Refresh widget by performing an immediate drawing.
    */
   protected void refresh()
   {
      Frame<?> frame = null;
      if (parent() == null || (frame = (Frame<?>) UiUtils.locateFrame(this)) == null)
         return;

      gd.drawFromOrigin(frame.getContentPane().screenPhysicalLocation(),  null, new Runnable()
      {
         @Override
         public void run()
         {
            draw();
            requestSync();
         }
      });
   }
   
   /**
    * Create the left-edge scroller.
    * 
    * @return   See above.
    */
   @Override
   protected EditScroller createLeftAnchor()
   {
      return new LeftAnchorEditScroller();
   }

   /**
    * Create the right-edge scroller.
    * 
    * @return   See above.
    */
   @Override
   protected EditScroller createRightAnchor()
   {
      return new RightAnchorEditScroller();
   }
   
   /**
    * Determine the portion of the widget's editable part.  This may be different than the widget's
    * width, as determined by the WIDTH-CHARS or WIDTH-PIXELS attribute.
    * <p>
    * For GUI widgets, this is the physical dimension without the insets.
    * 
    * @return   The editable dimension of this widget, in native units.
    */
   @Override
   protected NativeDimension editableNativeDimension()
   {
      int width = nativeWidth(this, gd, gf, false);
      
      NativeDimension dim = physicalDimension();
      // combo-box mode needs pure physical dimension based calculation
      if (comboBoxMode)
      {
//         dim.width -= nativeInsets.left + nativeInsets.right;
         dim.width -= 1;
      }
      else
      {
         dim.width = width - nativeInsets.right - nativeInsets.left;
      }
      dim.height-= nativeInsets.top + nativeInsets.bottom;
   
      return dim;
   }

   /**
    * Compute the location of the fill-in's editable content, relative to the fill-in's top-left
    * corner.
    * 
    * @return   See above.
    */
   private NativePoint editableNativeLocation()
   {
      NativePoint np = new NativePoint(nativeInsets.left, nativeInsets.top);
      
      return np;
   }

   /**
    * Compute the fill-in's content location, adjusted so that only the required portion will be
    * visible.
    * <p>
    * The returned location is relative to the fill-in's top-left corner, and the X coordinate
    * may be negative.  When used during drawing, the clipping rectangle ensures that nothing is
    * drawn outside of the required area.
    * 
    * @return   See above.
    */
   private NativePoint textNativeLocation()
   {
      String txt = getText(true);
      FontDetails<?> font = gf.font().font;

      NativePoint eloc = editableNativeLocation();
      NativeDimension edim = editableNativeDimension();
      boolean right = isRightAlignedInt();
      if (right)
      {
         // the text must be aligned on the rightmost column
         int txtWidth = gd.getTextWidth(txt, font);
         if (txtWidth < edim.width)
         {
            eloc.translate(edim.width - txtWidth, 0);
         }
      }
      
      NativePoint txtLoc = editScroller.adjustLocation(txt, eloc);

      return txtLoc;
   }

   /**
    * Compute the cursor's physical column, relative to the fill-in's leftside border.
    *  
    * @return   The x-axis location where the cursor needs to be drawn.
    */
   private int cursorPhysicalColumn()
   {
      FontCache fc = gf.font();
      FontDetails<?> font = fc.font;
   
      int cursorX = getNativeCursorPos();
      int offset = getCursorOffset();
      int last = getLastOffset();
      String txt = getText(true);
      int txtWidth = gd.getTextWidth(txt, font);
      int nwidth = editableNativeDimension().width;
      
      if (editScroller == rightAnchor && offset > last)
      {
         cursorX = nwidth;
      }
      
      if (offset == editScroller.getAnchorIndex())
      {
         if (editScroller == rightAnchor)
         {
            cursorX = nwidth;
         }
         else
         {
            if (isRightAlignedInt())
            {
               cursorX = nwidth < txtWidth ? 1 : nwidth - txtWidth + 1;
            }
            else
            {
               cursorX = 1;
            }
         }
      }
      else if (isRightAlignedInt())
      {
         String ctxt = offset > txt.length() ? null : txt.substring(offset);
         txtWidth = (ctxt == null ? 0 : gd.getTextWidth(ctxt, font));
         if (editScroller == rightAnchor)
         {
            cursorX = txtWidth > nwidth ? 1 : nwidth - txtWidth + 1;
         }
         else
         {
            cursorX = txtWidth < nwidth ? nwidth - txtWidth + 1 : cursorX;
         }
      }
      
      return cursorX;
   }

   /**
    * Move the cursor to the rightmost index.
    */
   private void moveCursorToLast()
   {
      setCursorOffset(getLastOffset() + 1);
   }

   /**
    * Move the cursor to the leftmost index.
    */
   private void moveCursorToFirst()
   {
      setCursorOffset(0);
   }

   /**
    * Content scroller implementation when the anchor is positioned on the left-side editable edge.
    */
   private class LeftAnchorEditScroller
   extends EditScroller
   {
      /**
       * Adjust the location so that only the visible content will be displayed.
       *  
       * @param    txt
       *           The full content.
       * @param    np
       *           The current location of the content.
       *           
       * @return   The adjusted location.
       */
      @Override
      public NativePoint adjustLocation(String txt, NativePoint np)
      {
         FontCache fc = gf.font();
         FontDetails<?> font = fc.font;
         
         int delta = gd.getTextWidth(txt.substring(0, Math.min(anchorIndex, txt.length())), font);
   
         return new NativePoint(np.x - delta, np.y);
      }
   
      /**
       * Get the amount of characters which are over the left-most edge, and are not being displayed.
       * 
       * @return   See above.
       */
      @Override
      public int getDelta()
      {
         return -anchorIndex;
      }
   
      /**
       * Get the index of the leftmost visible character in the widget's displayable area.
       * 
       * @return   See above.
       */
      @Override
      public int getOffset()
      {
         return anchorIndex;
      }
   }

   /**
    * Content scroller implementation when the anchor is positioned on the right-side editable edge.
    */
   private class RightAnchorEditScroller
   extends EditScroller
   {
      /**
       * Adjust the location so that only the visible content will be displayed.
       *  
       * @param    txt
       *           The full content.
       * @param    np
       *           The current location of the content.
       *           
       * @return   The adjusted location.
       */
      @Override
      public NativePoint adjustLocation(String txt, NativePoint np)
      {
         FontCache fc = gf.font();
         FontDetails<?> font = fc.font;
   
         int nwidth = editableNativeDimension().width;
         int twidth = gd.getTextWidth(txt.substring(0, Math.min(anchorIndex, txt.length())), font);
         int delta = nwidth - twidth - 1;
   
         return new NativePoint(np.x + delta, np.y);
      }
   
      /**
       * Get the amount of characters which are over the left-most edge, and are not being displayed.
       * 
       * @return   See above.
       */
      @Override
      public int getDelta()
      {
         return (int) width() - anchorIndex;
      }
      
      /**
       * Get the index of the leftmost visible character in the widget's displayable area.
       * 
       * @return   See above.
       */
      @Override
      public int getOffset()
      {
         return Math.max(anchorIndex - (int) width(), 0);
      }
   }
}
