/*
** Module   : ScrollPaneGuiImpl.java
** Abstract : scroll panel GUI implementation. 
**
** Copyright (c) 2014-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140917 First version.
** 002 MAG 20141016 Implements scroll bars management.
** 003 CA  20141123 The scrollbar visibility is off if they are AS_NEEDED.
** 004 CA  20141205 Do not use the frame's saved dimension, use the real one.
** 005 EVL 20150428 Adding method to set box flag.  Adding mouse handler for combo-box drop-down
**                  child implementation.
** 006 HC  20150508 Improvements to GUI window scrolling support.
** 007 CA  20150509 Added support for always hiding the scroll bars.  Some other misc fixes and
**                  improvements in preparation for GUI EDITOR widget.
** 008 EVL 20150511 Adding method to get the box flag.  Adding flag indicating srolling by mouse.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.event.*;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.ScrollBar.Bars;
import com.goldencode.p2j.ui.client.ScrollBar.Orientation;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * A GUI scroll pane implementation. 
 */
@SuppressWarnings("unchecked")
public class ScrollPaneGuiImpl
extends ScrollPane<GuiOutputManager>
{
   /**
    * Container show bars options:
    * 
    * AS_NEEDED - Only when needed.
    * ALWAYS    - Always
    */
   public enum ShowBars
   {
      AS_NEEDED,
      ALWAYS,
      NEVER
   }
   
   /** Screen driver */
   private GuiDriver gd;
   
   /** Color resolver */
   private GuiColorResolver gc;

   /** Horizontal scroll bar */
   private ScrollBarGuiImpl hBar;

   /** Vertical scroll bar */
   private ScrollBarGuiImpl vBar;
   
   /** Storage for listeners. */
   private ArrayList<ScrollListener> scrollListeners = new ArrayList<ScrollListener>();

   /** Scroll bars show policy */
   private ShowBars showBars = ShowBars.AS_NEEDED;
   
   /** True when have both horizontal and vertical scroll bars */
   boolean both;
   
   /** Draw box */
   boolean box;
   
   /**
    * Constructor.
    * 
    * @param   id
    *          The widget's ID.
    * @param   widget
    *          Child widget.
    */
   ScrollPaneGuiImpl(WidgetId id, Widget<GuiOutputManager> widget)
   {
      super(id, widget);
      // GUI driver
      gd = (GuiDriver) OutputManager.getDriver();
      gc = new GuiColorResolver(gd, config(), SysColor.COLOR_3DFACE, SysColor.COLOR_WINDOWTEXT);
      // widget inherit container width and height
      widget.setInheritWidth(true);
      widget.setInheritHeight(true);
      viewport().setExtent(0, 0);
      _setVisible(true);
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      boolean parentWidth = (isInheritWidth() && parent() != null);
      return (parentWidth) ? parent().width() : super.width();
   }

   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      boolean parentHeight = (isInheritHeight() && parent() != null);
      return (parentHeight) ? parent().height() : super.height();
   }
      
   /**
    * Adjust viewport size.
    */
   void adjustViewportSize()
   {
      Point origin = viewport().view().location();
      double w = width()  - barSize(Orientation.VERTICAL)   - origin.x;
      double h = height() - barSize(Orientation.HORIZONTAL) - origin.y;
            
      viewport().setSize(new Dimension(w, h));
   }
   
   /**
    * Widget drawing method.
    */
   @Override
   public void draw()
   {
      // draw only when widget is visible 
      if (!isVisible())
      {
         return;
      }

      // determine frame colors
      gc.colors();
      
      // resize viewport
      adjustViewportSize();
      
      adjustScrollBars();
      
      // get container coordinates
      final NativePoint p = physicalLocation();
      final NativeDimension d = physicalDimension();
            
      final NativeRectangle clip = clipRectangle(p, d);

      gd.draw(p, clip, new Runnable()
      {
         @Override
         public void run()
         {
            // draw container
            gd.setColor(gc.bgColor);
            gd.fillRect(0, 0, d.width, d.height);
            //gd.draw3DRect(0, 0, d.width, d.height, false);        

            // draw content pane
            viewport().draw();
                       
            // draw scroll bars
            drawScrollBars();
            
            // draw box
            if (box)
            {
               gd.setColor(gc.fgColor);
               gd.drawRect(0, 0, d.width - 1, d.height - 1); 
            }
         }
                  
         /**
          * Draw container scroll bars.
          */
         private void drawScrollBars()
         {
            if ((hBar != null) && hBar.isVisible())
            {
               hBar.wOff = 0;
            }
            if ((vBar != null) && vBar.isVisible())
            {
               vBar.hOff = 0;
            }               

            if (both)
            {
               // if at least one scroll bar is visible both will be visible
               // used by frames for example
               if (showBars == ShowBars.ALWAYS)
               {
                  boolean visible = hBar.isVisible() || vBar.isVisible();
                  hBar._setVisible(visible);
                  vBar._setVisible(visible);                  
               }
               
               // check if both scroll bars are visible
               boolean bothVisible = hBar.isVisible() && vBar.isVisible();
               if (bothVisible)
               {
                  NativeDimension hb = hBar.physicalDimension();
                  NativeDimension vb = vBar.physicalDimension(); 
                  // fill scroll bar bottom right intersection
                  int x = d.width  -  vb.width;
                  int y = d.height - hb.height; 
                  gd.setColor(gd.getSysColor(SysColor.COLOR_SCROLLBAR));
                  gd.fillRect(x, y, vb.width, hb.height);
                  hBar.wOff = vBar.width();
                  vBar.hOff = hBar.height();
               }
            }

            // draw horizontal scroll bar
            if ((hBar != null) && hBar.isVisible())
            {
               hBar.draw();
            }
            
            // draw vertical scroll bar
            if ((vBar != null) && vBar.isVisible())
            {
               vBar.draw();
            }            
         }         
      });
   }

   @Override
   public void draw(Widget<GuiOutputManager>[] list)
   {
      // TODO how to implement that?
   }
      
   /**
    * This method is invoked when scroll pane is scrolled.
    * Scroll events could coming from both scroll bars or from contained widget.
    * 
    * @param   event
    *          Event which triggered this call.
    */
   @Override
   public void scroll(ScrollEvent event)
   {      
      if (event.source() instanceof ScrollBar)
      {
         notify(viewport().view(), event);
      }
      else
      {
         // update scroll bars
         for (ScrollListener listener : scrollListeners)
         {
            listener.scroll(event);
         }
      }
   }
   
   /**
    * Set box flag indicating drawing frame around the scrollpane.
    * 
    * @param   flag
    *          New value for the box flag.
    */
   public void setBoxFlag(boolean flag)
   {
      box = flag;      
   }
   
   /**
    * Get box flag indicating drawing frame around the scrollpane.
    * 
    * @return  see above
    */
   public boolean getBoxFlag()
   {
      return box;      
   }
   
   /**
    * Set container scroll bars.
    * 
    * @param   bars
    *          Scroll bars.
    */
   public void setScrollBars(ScrollBar.Bars bars)
   {
      createScrollBars(bars);      
   }
   
   /**
    * Set show bars policy.
    * 
    * @param   showBars
    *          Show bars policy.
    */
   public void setShowBars(ShowBars showBars)
   {
      this.showBars = showBars;
   }
   
   /**
    *  Return horizontal scroll bar.
    *  
    * @return  see above.
    */
   ScrollBarGuiImpl hbar()
   {
      return hBar;
   }

   /**
    *  Return vertical scroll bar.
    *  
    * @return  see above.
    */
   public ScrollBarGuiImpl vbar()
   {
      return vBar;
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      if (child != null && child instanceof ScrollableSelectionListGuiImpl)
      {
         child.mousePressed(e);
      }
      else
      {
         super.mousePressed(e);
      }
   }

   /**
    * Get content pane.  Content pane MUST be a container instance.
    * 
    * @return  Content pane if view is a container <code>null</code> otherwise.
    */
   private AbstractContainer<GuiOutputManager> getContentPane()
   {
      Widget<GuiOutputManager> view = (Widget<GuiOutputManager>) viewport().view();
      
      // must be a container
      if (view instanceof AbstractContainer)
      {
         return (AbstractContainer<GuiOutputManager>) view;
      }      
      
      return null;
   }
   
   /**
    * Notify about scroll events.
    * 
    * @param   view
    *          Notified widget.
    * @param   event
    *          Scroll event.
    */
   private void notify(Widget<GuiOutputManager> view, ScrollEvent event)
   {
      if (view instanceof ScrollListener)
      {
         super.scroll(event);
      }
      else if (view instanceof Scrollable)
      {
         ScrollEvent evt = new ScrollEvent(this, event.direction(),
                                           event.limit(), event.mouseOrigin());
         ((Scrollable) view).processScrollEvent(evt);
      }               
   }
   
   /**
    * Set scroll bar visibility
    */
   private void adjustScrollBars()
   {
      AbstractContainer<GuiOutputManager> view = getContentPane();
      
      if (view == null)
      {
         return;
      }
 
      Dimension d, v;
      
      // special case for frames
      if (parent() instanceof Frame)
      {
         v = ((Frame<GuiOutputManager>) parent()).dimension();
         d = WindowWorkSpace.maxSize;         
      }
      else
      {
         d = dimension();
         v = view.getEffectiveScrollableSize();
      }
      
      // compute differences. multiply by 100 to get precision 
      int maxh = Coordinate.toInt((v.width  - d.width)  * 100.0f);
      int maxv = Coordinate.toInt((v.height - d.height) * 100.0f);         
            
      // horizontal
      adjustScrollBar(Orientation.HORIZONTAL, maxh, 100);
      if (hbar() != null)
      {
         // notify
         notify(view, new ScrollEvent(this, 
                                      ScrollDirection.LEFT, 
                                      new Point(0, hbar().getPosition())));
      }
      
      // vertical
      adjustScrollBar(Orientation.VERTICAL, maxv, 100);      
      if (vbar() != null)
      {
         // notify
         notify(view , new ScrollEvent(this, 
                                       ScrollDirection.UP, 
                                       new Point(0, vbar().getPosition())));
      }
   }
   
   /**
    * Adjust scroll bar max value.
    * 
    * @param   orientation
    *          Scroll bar orientation.
    * @param   max
    *          Max value.
    * @param   scale
    *          Scale factor.
    */
   private void adjustScrollBar(Orientation orientation, int max, int scale)
   {
      ScrollBarGuiImpl bar = (orientation == Orientation.HORIZONTAL) ? hBar : vBar;
      double bs = barSize(orientation == Orientation.HORIZONTAL ? Orientation.VERTICAL : 
                                                                  Orientation.HORIZONTAL);
      if (bar != null)
      {
         boolean visible = (max > 0) || this.showBars == ShowBars.ALWAYS;
         bar._setVisible(visible);
         if (visible)
         {
            bar.setMax(max / scale + Coordinate.toInt(bs));
         }
         else
         {
            if (parent() instanceof Frame && max < 0)
            {
               bar.setMax(Coordinate.toInt(bs));
            }
            else
            {
               bar.setMax(0);
            }
         }
         
         if (bar.getPosition() > bar.getMax())
         {
            bar.setPosition(bar.getMax());
         }
      }
   }
   
   /**
    * Get scroll bar size in characters.
    * 
    * @param   orientation
    *          Scroll bar orientation.
    *          
    * @return  see above
    */
   private double barSize(Orientation orientation)
   {
      double size = 0;
      
      switch (orientation)
      {
         case HORIZONTAL:
            if (hBar != null && hBar.isVisible())
            {
               size = hBar.height();
            }
            break;
         case VERTICAL:
            if (vBar != null && vBar.isVisible())
            {
               size = vBar.width();
            }
            break;
         default:
            break;
            
      }
      
      return size;
   }
         
   /**
    * Create and attach container scroll bars.
    * 
    * @param   bars
    *          Scroll bars.
    */
   private void createScrollBars(Bars bars)
   {
      switch (bars)
      {
         case NONE:
            break;
         case HORIZONTAL:
            if (hBar == null)
            {
               hBar = createScrollBar(Orientation.HORIZONTAL);
            }
            break;
         case VERTICAL:
            if (vBar == null)
            {
               vBar = createScrollBar(Orientation.VERTICAL);
            }
            break;
         case BOTH:
            both = true;
            if (hBar == null)
            {
               hBar = createScrollBar(Orientation.HORIZONTAL);
            }
            if (vBar == null)
            {
               vBar = createScrollBar(Orientation.VERTICAL);
            }
            break;
      }            
   }

   /**
    * Create scroll bar.
    * 
    * @param   orientation
    *          Scroll bar orientation.
    *          
    * @return  Scroll bar.
    */
   private ScrollBarGuiImpl createScrollBar(Orientation orientation)
   {
      ScrollBarGuiImpl bar = null;

      GuiWidgetFactory factory = screen().getFactory();
      bar = factory.createScrollBar(WidgetId.nextID(), orientation);        
      
      if (orientation == Orientation.HORIZONTAL)
      {
         bar.setInheritWidth(true);
      }
      else
      {
         bar.setInheritHeight(true);         
      }

      add(bar);
      bar.attach(this);
      bar.setVisible(showBars == ShowBars.ALWAYS || showBars == ShowBars.AS_NEEDED);
      scrollListeners.add(bar);
      
      bar.config().font = config().font;
      
      return bar;
   }
}
